from typing import *
import datetime
import re
import dateparser
import typing
import royalnet.commands as rc

from ..tables import MMEvent
from ..utils import MMTask


class MatchmakingCommand(rc.Command):
    name: str = "matchmaking"

    description: str = "Cerca persone per una partita a qualcosa!"

    syntax: str = ""

    aliases = ["mm", "lfg"]

    def __init__(self, interface: rc.CommandInterface):
        super().__init__(interface)

        # Find all active MMEvents and run the tasks for them
        session = self.alchemy.Session()

        # Create a new MMEvent and run it
        if self.interface.name == "telegram":
            MMEventT = self.alchemy.get(MMEvent)
            active_mmevents = (
                session
                .query(MMEventT)
                .filter(
                    MMEventT.interface == self.interface.name,
                    MMEventT.interrupted == False
                )
                .all()
            )
            for mmevent in active_mmevents:
                task = MMTask(mmevent.mmid, command=self)
                task.start()

    @staticmethod
    def _parse_args(args) -> Tuple[Optional[datetime.datetime], str, str]:
        """Parse command arguments, either using the standard syntax or the Proto syntax."""
        try:
            timestring, title, description = args.match(r"(?:\[\s*([^]]+)\s*]\s*)?([^\n]+)\s*\n?\s*(.+)?\s*", re.DOTALL)
        except rc.InvalidInputError:
            timestring, title, description = args.match(r"(?:\s*(.+?)\s*\n\s*)?([^\n]+)\s*\n?\s*(.+)?\s*", re.DOTALL)
        if timestring is not None:
            try:
                dt: typing.Optional[datetime.datetime] = dateparser.parse(timestring, settings={
                    "PREFER_DATES_FROM": "future"
                })
            except OverflowError:
                dt = None
            if dt is None:
                raise rc.InvalidInputError("La data che hai specificato non è valida.")
            if dt <= datetime.datetime.now():
                raise rc.InvalidInputError("La data che hai specificato è nel passato.")
            if dt - datetime.datetime.now() >= datetime.timedelta(days=366):
                raise rc.InvalidInputError("Hai specificato una data tra più di un anno!\n"
                                           "Se volevi scrivere un'orario, ricordati che le ore sono separate da "
                                           "due punti (:) e non da punto semplice!")
        else:
            dt = None
        return dt, title, description

    async def run(self, args: rc.CommandArgs, data: rc.CommandData) -> None:
        """Handle a matchmaking command call."""
        author = await data.get_author(error_if_none=True)

        # Parse the arguments, either with the standard syntax or with the Proto syntax
        dt, title, description = self._parse_args(args)

        # Add the MMEvent to the database
        mmevent: MMEvent = self.alchemy.get(MMEvent)(creator=author,
                                                     datetime=dt,
                                                     title=title,
                                                     description=description,
                                                     interface=self.interface.name)
        data.session.add(mmevent)
        await data.session_commit()

        # Create and run a task for the newly created MMEvent
        task = MMTask(mmevent.mmid, command=self)
        task.start()

        await data.reply(f"🚩 Matchmaking creato!")



from typing import *
import royalnet.commands as rc
import royalnet.backpack.tables as rbt

from ..tables import FiorygiTransaction


class MagickfiorygiCommand(rc.Command):
    name: str = "magickfiorygi"

    description: str = "Crea fiorygi dal nulla."

    syntax: str = "{destinatario} {quantità} {motivo}"

    async def run(self, args: rc.CommandArgs, data: rc.CommandData) -> None:
        author = await data.get_author(error_if_none=True)
        if "banker" not in author.roles:
            raise rc.UserError("Non hai permessi sufficienti per eseguire questo comando.")

        user_arg = args[0]
        qty_arg = args[1]
        reason_arg = " ".join(args[2:])

        if user_arg is None:
            raise rc.InvalidInputError("Non hai specificato un destinatario!")
        user = await rbt.User.find(self.alchemy, data.session, user_arg)
        if user is None:
            raise rc.InvalidInputError("L'utente specificato non esiste!")

        if qty_arg is None:
            raise rc.InvalidInputError("Non hai specificato una quantità!")
        try:
            qty = int(qty_arg)
        except ValueError:
            raise rc.InvalidInputError("La quantità specificata non è un numero!")

        if reason_arg == "":
            raise rc.InvalidInputError("Non hai specificato un motivo!")

        await FiorygiTransaction.spawn_fiorygi(data, user, qty, reason_arg)



import contextlib
import random
from typing import *
import logging
import datetime
import enum
import asyncio as aio

from telegram import InlineKeyboardMarkup as InKM
from telegram import InlineKeyboardButton as InKB
from telegram import Message as PTBMessage
from telegram import TelegramError

import royalnet.commands as rc
import royalnet.utils as ru
import royalnet.serf.telegram as rst

from ..types import MMChoice, MMInterfaceDataTelegram
from ..tables import MMEvent, MMResponse, FiorygiTransaction


class Interrupts(enum.Enum):
    TIME_RAN_OUT = enum.auto()
    MANUAL_START = enum.auto()
    MANUAL_DELETE = enum.auto()


log = logging.getLogger(__name__)


mmchoice_sorting = {
    MMChoice.YES: -4,
    MMChoice.LATE_SHORT: -3,
    MMChoice.LATE_MEDIUM: -2,
    MMChoice.LATE_LONG: -1,
    MMChoice.MAYBE: 0,
    MMChoice.NO: 1
}


class MMTask:
    def __init__(self, mmid: int, *, command: rc.Command):
        log.debug(f"Creating task for: {mmid}")

        self.loop: aio.AbstractEventLoop = command.loop
        self.task: Optional[aio.Task] = None
        self.queue: aio.Queue = aio.Queue(loop=self.loop)
        self.command: rc.Command = command
        self.mmid: int = mmid

        self._session: Optional = None
        self._EventT: Optional[Type[MMEvent]] = None
        self._ResponseT: Optional[Type[MMResponse]] = None
        self._mmevent: Optional[MMEvent] = None

    @property
    def is_running(self):
        return self.task is not None

    def sync(self):
        self._session.refresh(self._mmevent)

    def get_response_line(self, response: MMResponse):
        self.sync()

        # noinspection PyListCreation
        line = []

        # Emoji
        line.append(f"{response.choice.value}")

        # Mention the user if he said yes, otherwise just display his name
        if response.choice == MMChoice.NO:
            line.append(f"{response.user.telegram[0].name()}")
        else:
            line.append(f"{response.user.telegram[0].mention()}")

        # Late time
        if response.choice == MMChoice.LATE_SHORT:
            td = self._mmevent.datetime + datetime.timedelta(minutes=10)
            line.append(f"[{td.strftime('%H:%M')}]")

        elif response.choice == MMChoice.LATE_MEDIUM:
            td = self._mmevent.datetime + datetime.timedelta(minutes=30)
            line.append(f"[{td.strftime('%H:%M')}]")

        elif response.choice == MMChoice.LATE_LONG:
            td = self._mmevent.datetime + datetime.timedelta(minutes=60)
            line.append(f"[{td.strftime('%H:%M')}+]")

        # Creator
        if response.user == self._mmevent.creator:
            line.append("👑")

        # Result
        return " ".join(line)

    @property
    def channel_text(self) -> str:
        self.sync()

        # noinspection PyListCreation
        text = []

        # First line
        if self._mmevent.datetime is None:
            text.append(f"🌐 [Prossimamente] [b]{self._mmevent.title}[/b]")
        else:
            text.append(f"🚩 [{self._mmevent.datetime.strftime('%Y-%m-%d %H:%M')}] [b]{self._mmevent.title}[/b]")

        # Description
        if self._mmevent.description:
            text.append(f"{self._mmevent.description}")

        # Spacer
        text.append("")

        # Responses
        responses = sorted(self._mmevent.responses, key=lambda r: mmchoice_sorting[r.choice])
        for response in responses:
            text.append(self.get_response_line(response))

        # Result
        return "\n".join(text)

    @property
    def start_text(self) -> str:
        self.sync()

        # noinspection PyListCreation
        text = []

        # First line
        if self._mmevent.datetime is None:
            text.append(f"🌐 Le iscrizioni all'evento [b]{self._mmevent.title}[/b] sono terminate!")
        else:
            text.append(f"🚩 L'evento [b]{self._mmevent.title}[/b] è iniziato!")

        # Description
        if self._mmevent.description:
            text.append(f"{self._mmevent.description}")

        # Spacer
        text.append("")

        # Responses
        responses = sorted(self._mmevent.responses, key=lambda r: mmchoice_sorting[r.choice])
        for response in responses:
            text.append(self.get_response_line(response))

        # Result
        return "\n".join(text)

    @property
    def delete_text(self) -> str:
        return f"🗑 L'evento [b]{self._mmevent.title}[/b] è stato eliminato."

    def get_answer_callback(self, choice: MMChoice):
        async def callback(data: rc.CommandData):
            # Find the user who clicked on the button
            user = await data.get_author(error_if_none=True)

            # Get the related MMEvent
            mmevent: MMEvent = await ru.asyncify(data.session.query(self._EventT).get, self.mmid)

            # Check if the user had already responded
            mmresponse: MMResponse = await ru.asyncify(
                data.session.query(self._ResponseT).filter_by(user=user, mmevent=mmevent).one_or_none
            )

            if mmresponse is None:
                # If they didn't respond, create a new MMResponse
                mmresponse = self._ResponseT(user=user, mmevent=mmevent, choice=choice)
                data.session.add(mmresponse)

                # Drop fiorygi
                if random.randrange(100) < self.command.config["Matchmaking"]["fiorygi_award_chance"]:
                    await FiorygiTransaction.spawn_fiorygi(data, user, 1, "aver risposto a un matchmaking")
            else:
                # Change their response
                mmresponse.choice = choice
            await data.session_commit()

            await self.telegram_channel_message_update()

            await data.reply(f"{choice.value} Hai risposto al matchmaking!")
        return callback

    def get_delete_callback(self):
        async def callback(data: rc.CommandData):
            # Find the user who clicked on the button
            user = await data.get_author(error_if_none=True)

            # Get the related MMEvent
            mmevent: MMEvent = await ru.asyncify(data.session.query(self._EventT).get, self.mmid)

            # Ensure the user has the required roles to start the matchmaking
            if user != mmevent.creator and "admin" not in user.roles:
                raise rc.UserError("Non hai i permessi per eliminare questo matchmaking!")

            # Interrupt the matchmaking with the MANUAL_DELETE reason
            await self.queue.put(Interrupts.MANUAL_DELETE)

            await data.reply(f"🗑 Evento eliminato!")
        return callback

    def get_start_callback(self):
        async def callback(data: rc.CommandData):
            # Find the user who clicked on the button
            user = await data.get_author(error_if_none=True)

            # Get the related MMEvent
            mmevent: MMEvent = await ru.asyncify(data.session.query(self._EventT).get, self.mmid)

            # Ensure the user has the required roles to start the matchmaking
            if user != mmevent.creator and "admin" not in user.roles:
                raise rc.UserError("Non hai i permessi per eliminare questo matchmaking!")

            # Interrupt the matchmaking with the MANUAL_DELETE reason
            await self.queue.put(Interrupts.MANUAL_START)

            await data.reply(f"🚩 Evento avviato!")
        return callback

    @property
    def royalnet_keyboard(self):
        # noinspection PyListCreation
        rows = []

        rows.append([
            rc.KeyboardKey(
                interface=self.command.interface,
                short=f"{MMChoice.YES.value}",
                text="Ci sarò!",
                callback=self.get_answer_callback(MMChoice.YES)
            ),
            rc.KeyboardKey(
                interface=self.command.interface,
                short=f"{MMChoice.MAYBE.value}",
                text="Forse...",
                callback=self.get_answer_callback(MMChoice.MAYBE)
            ),
            rc.KeyboardKey(
                interface=self.command.interface,
                short=f"{MMChoice.NO.value}",
                text="Non mi interessa.",
                callback=self.get_answer_callback(MMChoice.NO)
            ),
        ])

        if self._mmevent.datetime is not None:
            rows.append([
                rc.KeyboardKey(
                    interface=self.command.interface,
                    short=f"{MMChoice.LATE_SHORT.value}",
                    text="10 min",
                    callback=self.get_answer_callback(MMChoice.LATE_SHORT)
                ),
                rc.KeyboardKey(
                    interface=self.command.interface,
                    short=f"{MMChoice.LATE_MEDIUM.value}",
                    text="30 min",
                    callback=self.get_answer_callback(MMChoice.LATE_MEDIUM)
                ),
                rc.KeyboardKey(
                    interface=self.command.interface,
                    short=f"{MMChoice.LATE_LONG.value}",
                    text="60 min",
                    callback=self.get_answer_callback(MMChoice.LATE_LONG)
                )
            ])

        rows.append([
            rc.KeyboardKey(
                interface=self.command.interface,
                short=f"🗑",
                text="Elimina",
                callback=self.get_delete_callback()
            ),
            rc.KeyboardKey(
                interface=self.command.interface,
                short=f"🚩",
                text="Inizia",
                callback=self.get_start_callback()
            ),
        ])

        return rows

    @property
    def telegram_keyboard(self):
        # noinspection PyListCreation
        rows = []
        key_id = 0

        for r_row in self.royalnet_keyboard:
            row = []
            for r_key in r_row:
                # Generate a unique callback string
                callback_str = f"mm{self.mmid}_{key_id}"

                # Create a InlineKeyboardButton with that callback string
                row.append(InKB(f"{r_key.short} {r_key.text}", callback_data=callback_str))

                # Increase the key_id
                key_id += 1
            rows.append(row)

        # Return the resulting InlineKeyboardMarkup
        return InKM(rows)

    def register_telegram_keyboard(self, inkm: InKM):
        # noinspection PyListCreation
        royalnet_keyboard = self.royalnet_keyboard
        for x, row in enumerate(inkm.inline_keyboard):
            for y, key in enumerate(row):
                key: InKB
                self.command.interface.serf.register_keyboard_key(key.callback_data, key=royalnet_keyboard[x][y])

    def unregister_telegram_keyboard(self, inkm: InKM):
        for row in inkm.inline_keyboard:
            for key in row:
                key: InKB
                self.command.interface.serf.unregister_keyboard_key(key.callback_data)

    async def wait_until_due(self):
        """When the event is due, interrupt the MMTask with the TIME_RAN_OUT reason."""
        if self._mmevent.datetime is None:
            return
        await ru.sleep_until(self._mmevent.datetime)
        await self.queue.put(Interrupts.TIME_RAN_OUT)

    @property
    def telegram_channel_id(self):
        return self.command.config["Matchmaking"]["mm_telegram_channel_id"]

    @property
    def telegram_group_id(self):
        return self.command.config["Matchmaking"]["mm_telegram_group_id"]

    @contextlib.asynccontextmanager
    async def telegram_channel_message(self):

        # Generate the InlineKeyboardMarkup
        inkm = self.telegram_keyboard

        # Bind the Royalnet buttons to the Telegram keyboard
        log.debug(f"Registering keyboard for: {self.mmid}")
        self.register_telegram_keyboard(inkm)

        # If the event has no associated interface data...
        if self._mmevent.interface_data is None:
            # Send the channel message
            log.debug(f"Sending message for: {self.mmid}")
            message: PTBMessage = await self.command.interface.serf.api_call(
                self.command.interface.serf.client.send_message,
                chat_id=self.telegram_channel_id,
                text=rst.escape(self.channel_text),
                parse_mode="HTML",
                disable_webpage_preview=True,
                reply_markup=inkm
            )

            # Register the interface data on the database
            self._mmevent.interface_data = MMInterfaceDataTelegram(
                chat_id=self.telegram_channel_id,
                message_id=message.message_id
            )
            self._session.commit()

        # Wait until the event starts
        yield

        # Delete the channel message
        log.debug(f"Deleting message for: {self.mmid}")
        await self.command.interface.serf.api_call(
            self.command.interface.serf.client.delete_message,
            chat_id=self._mmevent.interface_data.chat_id,
            message_id=self._mmevent.interface_data.message_id
        )

        # Unregister the Telegram keyboard bindings
        log.debug(f"Unregistering keyboard for: {self.mmid}")
        self.unregister_telegram_keyboard(inkm)

    async def telegram_channel_message_update(self):
        log.debug(f"Updating message for: {self.mmid}")
        try:
            await ru.asyncify(
                self.command.interface.serf.client.edit_message_text,
                chat_id=self._mmevent.interface_data.chat_id,
                text=rst.escape(self.channel_text),
                message_id=self._mmevent.interface_data.message_id,
                parse_mode="HTML",
                disable_web_page_preview=True,
                reply_markup=self.telegram_keyboard
            )
        except TelegramError as e:
            log.warning(f"TelegramError during update: {e}")

    async def telegram_group_message_start(self):
        await self.command.interface.serf.api_call(
            self.command.interface.serf.client.send_message,
            chat_id=self.telegram_group_id,
            text=rst.escape(self.start_text),
            parse_mode="HTML",
            disable_webpage_preview=True
        )

    async def telegram_group_message_delete(self):
        await self.command.interface.serf.api_call(
            self.command.interface.serf.client.send_message,
            chat_id=self.telegram_group_id,
            text=rst.escape(self.delete_text),
            parse_mode="HTML",
            disable_webpage_preview=True
        )

    def start(self):
        log.debug(f"Starting task for: {self.mmid}")
        self.task = self.loop.create_task(self.run())

    @ru.sentry_async_wrap()
    async def run(self):
        log.debug(f"Running task for: {self.mmid}")

        # Create a new session for the MMTask
        self._session = self.command.alchemy.Session()
        self._EventT = self.command.alchemy.get(MMEvent)
        self._ResponseT = self.command.alchemy.get(MMResponse)
        self._mmevent: MMEvent = self._session.query(self._EventT).get(self.mmid)

        if self._mmevent is None:
            raise rc.InvalidInputError(f"No event exists with the mmid {self.mmid}.")

        if self._mmevent.interface != "telegram":
            raise rc.UnsupportedError("Currently only the Telegram interface is supported.")

        async with self.telegram_channel_message():
            self.loop.create_task(self.wait_until_due())

            # Sleep until something interrupts the task
            interrupt = await self.queue.get()

            # Mark the event as interrupted
            self._mmevent.interrupted = True
            self._session.commit()

        # Send a group notification if the MMEvent wasn't deleted
        if interrupt != Interrupts.MANUAL_DELETE:
            await self.telegram_group_message_start()
        else:
            await self.telegram_group_message_delete()

        # Close the database session
        await ru.asyncify(self._session.close)


# Imports go here!
from .api_bio import ApiBioSetStar
from .api_diario import ApiDiarioGetStar
from .api_diario_list import ApiDiarioPagesStar
from .api_discord_cv import ApiDiscordCvStar
from .api_discord_play import ApiDiscordPlayStar
from .api_fiorygi import ApiFiorygiStar
from .api_diario_random import ApiDiarioRandomStar
from .api_poll import ApiPollStar
from .api_poll_list import ApiPollsListStar
from .api_cvstats_latest import ApiCvstatsLatestStar
from .api_cvstats_avg import ApiCvstatsAvgStar
from .api_user_ryg import ApiUserRygStar
from .api_user_ryg_list import ApiUserRygListStar
from .api_user_avatar import ApiUserAvatarStar
from .api_auth_login_osu import ApiAuthLoginOsuStar

# Enter the PageStars of your Pack here!
available_page_stars = [
    ApiBioSetStar,
    ApiDiarioGetStar,
    ApiDiarioPagesStar,
    ApiDiscordCvStar,
    ApiDiscordPlayStar,
    ApiFiorygiStar,
    ApiDiarioRandomStar,
    ApiPollStar,
    ApiPollsListStar,
    ApiCvstatsLatestStar,
    ApiCvstatsAvgStar,
    ApiUserRygStar,
    ApiUserRygListStar,
    ApiUserAvatarStar,
    ApiAuthLoginOsuStar,
]

# Don't change this, it should automatically generate __all__
__all__ = [star.__name__ for star in available_page_stars]



import enum


class LeagueRank(enum.Enum):
    I = 1
    II = 2
    III = 3
    IV = 4

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"{self.__class__.__qualname__}.{self.name}"

    def __gt__(self, other):
        return self.value < other.value

    @classmethod
    def from_string(cls, string: str):
        return cls.__members__.get(string)



from typing import *
import royalnet.commands as rc
import aiohttp
import io


class DogCommand(rc.Command):
    name: str = "dog"

    description: str = "Invia un cane della razza specificata in chat."

    syntax: str = "[razza|list]"

    _breeds = [
        "affenpinscher",
        "african",
        "airedale",
        "akita",
        "appenzeller",
        "australian-shepherd",
        "basenji",
        "beagle",
        "bluetick",
        "borzoi",
        "bouvier",
        "boxer",
        "brabancon",
        "briard",
        "buhund-norwegian",
        "bulldog-boston",
        "bulldog-english",
        "bulldog-french",
        "bullterrier-staffordshire",
        "cairn",
        "cattledog-australian",
        "chihuahua",
        "chow",
        "clumber",
        "cockapoo",
        "collie-border",
        "coonhound",
        "corgi-cardigan",
        "cotondetulear",
        "dachshund",
        "dalmatian",
        "dane-great",
        "deerhound-scottish",
        "dhole",
        "dingo",
        "doberman",
        "elkhound-norwegian",
        "entlebucher",
        "eskimo",
        "finnish-lapphund",
        "frise-bichon",
        "germanshepherd",
        "greyhound-italian",
        "groenendael",
        "havanese",
        "hound-afghan",
        "hound-basset",
        "hound-blood",
        "hound-english",
        "hound-ibizan",
        "hound-plott",
        "hound-walker",
        "husky",
        "keeshond",
        "kelpie",
        "komondor",
        "kuvasz",
        "labrador",
        "leonberg",
        "lhasa",
        "malamute",
        "malinois",
        "maltese",
        "mastiff-bull",
        "mastiff-english",
        "mastiff-tibetan",
        "mexicanhairless",
        "mix",
        "mountain-bernese",
        "mountain-swiss",
        "newfoundland",
        "otterhound",
        "ovcharka-caucasian",
        "papillon",
        "pekinese",
        "pembroke",
        "pinscher-miniature",
        "pitbull",
        "pointer-german",
        "pointer-germanlonghair",
        "pomeranian",
        "poodle-miniature",
        "poodle-standard",
        "poodle-toy",
        "pug",
        "puggle",
        "pyrenees",
        "redbone",
        "retriever-chesapeake",
        "retriever-curly",
        "retriever-flatcoated",
        "retriever-golden",
        "ridgeback-rhodesian",
        "rottweiler",
        "saluki",
        "samoyed",
        "schipperke",
        "schnauzer-giant",
        "schnauzer-miniature",
        "setter-english",
        "setter-gordon",
        "setter-irish",
        "sheepdog-english",
        "sheepdog-shetland",
        "shiba",
        "shihtzu",
        "spaniel-blenheim",
        "spaniel-brittany",
        "spaniel-cocker",
        "spaniel-irish",
        "spaniel-japanese",
        "spaniel-sussex",
        "spaniel-welsh",
        "springer-english",
        "stbernard",
        "terrier-american",
        "terrier-australian",
        "terrier-bedlington",
        "terrier-border",
        "terrier-dandie",
        "terrier-fox",
        "terrier-irish",
        "terrier-kerryblue",
        "terrier-lakeland",
        "terrier-norfolk",
        "terrier-norwich",
        "terrier-patterdale",
        "terrier-russell",
        "terrier-scottish",
        "terrier-sealyham",
        "terrier-silky",
        "terrier-tibetan",
        "terrier-toy",
        "terrier-westhighland",
        "terrier-wheaten",
        "terrier-yorkshire",
        "vizsla",
        "waterdog-spanish",
        "weimaraner",
        "whippet",
        "wolfhound-irish",
    ]

    async def run(self, args: rc.CommandArgs, data: rc.CommandData) -> None:
        breed = args.joined()
        if breed:
            if breed == "list":
                await data.reply("\n".join(["ℹ️ Razze disponibili:", [f"[c]{breed}[/c]" for breed in self._breeds]]))
            if breed in self._breeds:
                url = f"https://dog.ceo/api/breed/{breed}/images/random"
            else:
                raise rc.InvalidInputError("Questa razza non è disponibile.\n")
        else:
            url = f"https://dog.ceo/api/breeds/image/random"

        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                if response.status >= 400:
                    raise rc.ExternalError(f"Request returned {response.status}")
                result = await response.json()
                assert "status" in result
                assert result["status"] == "success"
                assert "message" in result
                url = result["message"]
            async with session.get(url) as response:
                img = await response.content.read()
                await data.reply_image(image=io.BytesIO(img))



import logging
import telegram
from typing import *
from royalnet.serf.telegram.telegramserf import TelegramSerf, escape
from royalnet.commands import *


log = logging.getLogger(__name__)


class TelegramMessageEvent(Event):
    name = "telegram_message"

    async def run(self, chat_id, text, **kwargs) -> dict:
        if not self.interface.name == "telegram":
            raise UnsupportedError()

        # noinspection PyTypeChecker
        serf: TelegramSerf = self.interface.serf

        log.debug("Forwarding message from Herald to Telegram.")
        await serf.api_call(serf.client.send_message,
                            chat_id=chat_id,
                            text=escape(text),
                            parse_mode="HTML",
                            disable_web_page_preview=True)

        return {}



import royalnet.utils as ru
import royalnet.constellation.api as rca


class ApiDiscordCvStar(rca.ApiStar):
    path = "/api/discord/cv/v1"

    tags = ["discord"]

    @rca.magic
    async def get(self, data: rca.ApiData) -> ru.JSON:
        """Get the members status of a single Discord guild.

        Equivalent to calling /cv in a chat."""
        response = await self.interface.call_herald_event("discord", "discord_cv")
        return response



from typing import *
import royalnet.constellation.api as rca
import royalnet.utils as ru
from ..tables import *
from sqlalchemy import func


class ApiDiarioRandomStar(rca.ApiStar):
    path = "/api/diario/random/v1"

    parameters = {
        "get": {
            "amount": "The number of diario entries to get."
        }
    }

    tags = ["diario"]

    @rca.magic
    async def get(self, data: rca.ApiData) -> ru.JSON:
        """Get random diario entries."""
        DiarioT = self.alchemy.get(Diario)
        try:
            amount = int(data["amount"])
        except ValueError:
            raise rca.InvalidParameterError("'amount' is not a valid int.")
        entries: List[Diario] = await ru.asyncify(
            data.session
                .query(DiarioT)
                .order_by(func.random())
                .limit(amount)
                .all
        )
        if len(entries) < amount:
            raise rca.NotFoundError("Not enough diario entries.")
        return list(map(lambda e: e.json(), entries))



from sqlalchemy import *
from sqlalchemy.orm import relationship, backref
from sqlalchemy.ext.declarative import declared_attr


class Fiorygi:
    __tablename__ = "fiorygi"

    @declared_attr
    def user_id(self):
        return Column(Integer, ForeignKey("users.uid"), primary_key=True)

    @declared_attr
    def user(self):
        return relationship("User", backref=backref("fiorygi", uselist=False))

    @declared_attr
    def fiorygi(self):
        return Column(Integer, nullable=False, default=0)

    def __repr__(self):
        return f"<{self.__class__.__name__} di {self.user}: {self.fiorygi}>"

    def __str__(self):
        return f"{self.fiorygi} fioryg" + ("i" if self.fiorygi != 1 else "")



# Imports go here!
from .diario import Diario
from .wikipages import WikiPage
from .bios import Bio
from .reminders import Reminder
from .triviascores import TriviaScore
from .leagueoflegends import LeagueOfLegends
from .fiorygi import Fiorygi
from .steam import Steam
from .dota import Dota
from .fiorygitransactions import FiorygiTransaction
from .brawlhalla import Brawlhalla
from .polls import Poll
from .pollcomments import PollComment
from .pollvotes import PollVote
from .brawlhalladuos import BrawlhallaDuo
from .mmevents import MMEvent
from .mmresponse import MMResponse
from .cvstats import Cvstats
from .treasure import Treasure
from .osu import Osu

# Enter the tables of your Pack here!
available_tables = [
    Diario,
    WikiPage,
    Bio,
    Reminder,
    TriviaScore,
    LeagueOfLegends,
    Fiorygi,
    Steam,
    Dota,
    FiorygiTransaction,
    Brawlhalla,
    Poll,
    PollComment,
    PollVote,
    BrawlhallaDuo,
    MMEvent,
    MMResponse,
    Cvstats,
    Treasure,
    Osu,
]

# Don't change this, it should automatically generate __all__
__all__ = [table.__name__ for table in available_tables]



import royalnet.backpack.tables as rbt
import royalnet.constellation.api as rca


class ApiUserRygStar(rca.ApiStar):
    path = "/api/user/ryg/v2"

    parameters = {
        "get": {
            "uid": "(Choose one) The id of the user to get information about.",
            "alias": "(Choose one) The alias of the user to get information about.",
        }
    }

    tags = ["user"]

    async def get_user(self, data: rca.ApiData):
        uid = data.int("uid", optional=True)
        alias = data.str("alias", optional=True)

        if uid:
            user = await rbt.User.find(self.alchemy, data.session, uid)
        elif alias:
            user = await rbt.User.find(self.alchemy, data.session, alias)
        else:
            raise rca.MissingParameterError("Neither uid or alias were specified.")

        if user is None:
            raise rca.NotFoundError("No such user.")

        return user

    @rca.magic
    async def get(self, data: rca.ApiData) -> dict:
        """Get Royalpack information about a user."""
        user = await self.get_user(data)
        result = {
            **user.json(),
            "bio": user.bio.json() if user.bio is not None else None,
            "fiorygi": user.fiorygi.fiorygi if user.fiorygi is not None else None,
            "steam": [steam.json() for steam in user.steam],
            "leagueoflegends": [leagueoflegends.json() for leagueoflegends in user.leagueoflegends],
            "trivia": user.trivia_score.json() if user.trivia_score is not None else None
        }
        return result



from typing import *
import logging
import aiohttp
import royalnet.commands as rc
import royalnet.utils as ru
from royalnet.backpack import tables as rbt
from .abstract.linker import LinkerCommand

from ..tables import Steam, Dota
from ..types import DotaRank

log = logging.getLogger(__name__)


class DotaCommand(LinkerCommand):
    name: str = "dota"

    aliases = ["dota2", "doto", "doto2", "dotka", "dotka2"]

    description: str = "Visualizza le tue statistiche di Dota."

    syntax: str = ""

    def describe(self, obj: Steam) -> str:
        string = f"ℹ️ [b]{obj.persona_name}[/b]\n"
        if obj.dota.rank:
            string += f"{obj.dota.rank}\n"
        string += f"\n" \
                  f"Wins: [b]{obj.dota.wins}[/b]\n" \
                  f"Losses: [b]{obj.dota.losses}[/b]\n" \
                  f"\n"
        return string

    async def get_updatables_of_user(self, session, user: rbt.User) -> List[Dota]:
        return user.steam

    async def get_updatables(self, session) -> List[Dota]:
        return await ru.asyncify(session.query(self.alchemy.get(Steam)).all)

    async def create(self,
                     session,
                     user: rbt.User,
                     args: rc.CommandArgs,
                     data: Optional[rc.CommandData] = None) -> Optional[Dota]:
        raise rc.InvalidInputError("Dota accounts are automatically linked from Steam.")

    async def update(self, session, obj: Steam, change: Callable[[str, Any], Awaitable[None]]):
        log.debug(f"Getting player data from OpenDota...")
        async with aiohttp.ClientSession() as hcs:
            # Get profile data
            async with hcs.get(f"https://api.opendota.com/api/players/{obj.steamid.as_32}/") as response:
                if response.status != 200:
                    raise rc.ExternalError(f"OpenDota / returned {response.status}!")
                p = await response.json()
                # No such user
                if "profile" not in p:
                    log.debug(f"Not found: {obj}")
                    return
            # Get win/loss data
            async with hcs.get(f"https://api.opendota.com/api/players/{obj.steamid.as_32}/wl") as response:
                if response.status != 200:
                    raise rc.ExternalError(f"OpenDota /wl returned {response.status}!")
                wl = await response.json()
                # No such user
                if wl["win"] == 0 and wl["lose"] == 0:
                    log.debug(f"Not found: {obj}")
                    return
        # Find the Dota record, if it exists
        dota: Dota = obj.dota
        if dota is None:
            # Autocreate the Dota record
            dota = self.alchemy.get(Dota)(steam=obj)
            session.add(dota)
            session.flush()

        # Make a custom change function
        async def change(attribute: str, new: Any):
            await self._change(session=session, obj=dota, attribute=attribute, new=new)

        await change("wins", wl["win"])
        await change("losses", wl["lose"])
        if p["rank_tier"]:
            await change("rank", DotaRank(rank_tier=p["rank_tier"]))
        else:
            await change("rank", None)

    async def on_increase(self, session, obj: Dota, attribute: str, old: Any, new: Any) -> None:
        if attribute == "rank":
            await self.notify(f"📈 [b]{obj.steam.user}[/b] è salito a [b]{new}[/b] su Dota 2! Congratulazioni!")

    async def on_unchanged(self, session, obj: Dota, attribute: str, old: Any, new: Any) -> None:
        pass

    async def on_decrease(self, session, obj: Dota, attribute: str, old: Any, new: Any) -> None:
        if attribute == "rank":
            await self.notify(f"📉 [b]{obj.steam.user}[/b] è sceso a [b]{new}[/b] su Dota 2.")

    async def on_first(self, session, obj: Dota, attribute: str, old: None, new: Any) -> None:
        if attribute == "wins":
            await self.notify(f"↔️ Account {obj} connesso a {obj.steam.user}!")
        elif attribute == "rank":
            await self.notify(f"🌟 [b]{obj.steam.user}[/b] si è classificato [b]{new}[/b] su Dota 2!")

    async def on_reset(self, session, obj: Dota, attribute: str, old: Any, new: None) -> None:
        if attribute == "rank":
            await self.notify(f"⬜️ [b]{obj.steam.user}[/b] non ha più un rank su Dota 2.")



import pickle
from sqlalchemy import *
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr


class MMEvent:
    __tablename__ = "mmevents"

    @declared_attr
    def creator_id(self):
        return Column(Integer, ForeignKey("users.uid"), nullable=False)

    @declared_attr
    def creator(self):
        return relationship("User", backref="mmevents_created")

    @declared_attr
    def mmid(self):
        return Column(Integer, primary_key=True)

    @declared_attr
    def datetime(self):
        return Column(DateTime)

    @declared_attr
    def title(self):
        return Column(String, nullable=False)

    @declared_attr
    def description(self):
        return Column(Text, nullable=False, default="")

    @declared_attr
    def interface(self):
        return Column(String, nullable=False)

    @declared_attr
    def raw_interface_data(self):
        # The default is a pickled None
        return Column(Binary, nullable=False, default=b'\x80\x03N.')

    @property
    def interface_data(self):
        return pickle.loads(self.raw_interface_data)

    @interface_data.setter
    def interface_data(self, value):
        self.raw_interface_data = pickle.dumps(value)

    @declared_attr
    def interrupted(self):
        return Column(Boolean, nullable=False, default=False)

    def __repr__(self):
        return f"<MMEvent {self.mmid}: {self.title}>"



import royalnet.constellation.api as rca
import royalnet.utils as ru
from ..tables import *


class ApiDiarioGetStar(rca.ApiStar):
    path = "/api/diario/v2"

    parameters = {
        "get": {
            "id": "The id of the diario entry to get."
        }
    }

    tags = ["diario"]

    @rca.magic
    async def get(self, data: rca.ApiData) -> ru.JSON:
        """Get a specific diario entry."""
        diario_id = data.int("id")
        entry: Diario = await ru.asyncify(data.session.query(self.alchemy.get(Diario)).get, diario_id)
        if entry is None:
            raise rca.NotFoundError("No such diario entry.")
        return entry.json()



from typing import *
import datetime
import uuid
import royalnet.utils as ru
import royalnet.constellation.api as rca
from ..tables import Poll


class ApiPollStar(rca.ApiStar):
    path = "/api/poll/v2"

    parameters = {
        "get": {
            "uuid": "The UUID of the poll to get.",
        },
        "post": {
            "question": "The question to ask in the poll.",
            "description": "A longer Markdown-formatted description.",
            "expires": "A ISO timestamp of the expiration date for the poll.",
        }
    }

    auth = {
        "get": False,
        "post": True
    }

    tags = ["poll"]

    @rca.magic
    async def get(self, data: rca.ApiData) -> ru.JSON:
        """Get a specific poll."""
        PollT = self.alchemy.get(Poll)

        try:
            pid = uuid.UUID(data["uuid"])
        except (ValueError, AttributeError, TypeError):
            raise rca.InvalidParameterError("'uuid' is not a valid UUID.")

        poll: Poll = await ru.asyncify(data.session.query(PollT).get, pid)
        if poll is None:
            raise rca.NotFoundError("No such page.")

        return poll.json()

    @rca.magic
    async def post(self, data: rca.ApiData) -> ru.JSON:
        """Create a new poll."""
        PollT = self.alchemy.get(Poll)

        poll = PollT(
            id=uuid.uuid4(),
            creator=await data.user(),
            created=datetime.datetime.now(),
            expires=datetime.datetime.fromisoformat(data["expires"]) if "expires" in data else None,
            question=data["question"],
            description=data.get("description"),
        )

        data.session.add(poll)
        await data.session_commit()

        return poll.json()



from .dotamedal import DotaMedal
from .dotastars import DotaStars


class DotaRank:
    __slots__ = "medal", "stars"

    def __init__(self, medal: DotaMedal = None, stars: DotaStars = None, *, rank_tier: int = None):
        if rank_tier is not None:
            self.medal: DotaMedal = DotaMedal(rank_tier // 10)
            self.stars: DotaStars = DotaStars(rank_tier % 10)
        else:
            if medal is None or stars is None:
                raise AttributeError("Missing medal, stars or rank_tier.")
            self.medal = medal
            self.stars = stars

    def __gt__(self, other):
        if other is None:
            return True
        if not isinstance(other, DotaRank):
            raise TypeError(f"Can't compare {self.__class__.__qualname__} with {other.__class__.__qualname__}")
        if self.medal > other.medal:
            return True
        elif self.medal < other.medal:
            return False
        elif self.stars > other.stars:
            return True
        return False

    def __eq__(self, other):
        if other is None:
            return False
        if not isinstance(other, DotaRank):
            raise TypeError(f"Can't compare {self.__class__.__qualname__} with {other.__class__.__qualname__}")
        return self.medal == other.medal and self.stars == other.stars

    def __repr__(self):
        return f"<{self.__class__.__qualname__}: {self.medal.name} {self.stars.name}>"

    def __str__(self):
        return f"{self.medal} {self.stars}"

    @property
    def rank_tier(self) -> int:
        return (self.medal.value * 10 + self.stars.value)


from typing import *
import royalnet.commands as rc
import royalnet.utils as ru
from sqlalchemy import func

from ..tables import Diario


class DiarioshuffleCommand(rc.Command):
    name: str = "diarioshuffle"

    description: str = "Cita una riga casuale del diario."

    aliases = ["dis", "dishuffle", "dish"]

    syntax = ""

    async def run(self, args: rc.CommandArgs, data: rc.CommandData) -> None:
        DiarioT = self.alchemy.get(Diario)
        entry: List[Diario] = await ru.asyncify(
            data.session
                .query(DiarioT)
                .order_by(func.random())
                .limit(1)
                .one_or_none
        )
        if entry is None:
            raise rc.CommandError("Nessuna riga del diario trovata.")
        await data.reply(f"ℹ️ {entry}")



from sqlalchemy import *
from sqlalchemy.orm import *
from sqlalchemy.ext.declarative import declared_attr


class Cvstats:
    __tablename__ = "cvstats"

    @declared_attr
    def id(self):
        return Column(Integer, primary_key=True)

    @declared_attr
    def timestamp(self):
        return Column(DateTime)

    @declared_attr
    def members_connected(self):
        return Column(Integer)

    @declared_attr
    def users_connected(self):
        return Column(Integer)

    @declared_attr
    def members_online(self):
        return Column(Integer)

    @declared_attr
    def users_online(self):
        return Column(Integer)

    @declared_attr
    def members_playing(self):
        return Column(Integer)

    @declared_attr
    def users_playing(self):
        return Column(Integer)

    @declared_attr
    def members_total(self):
        return Column(Integer)

    @declared_attr
    def users_total(self):
        return Column(Integer)

    def json(self):
        return {
            "timestamp": self.timestamp.isoformat(),
            "users_total": self.users_total,
            "members_total": self.members_total,
            "users_online": self.users_online,
            "members_online": self.members_online,
            "users_connected": self.users_connected,
            "members_connected": self.members_connected,
            "users_playing": self.users_playing,
            "members_playing": self.members_playing,
        }


from typing import *
import steam.steamid
import steam.webapi
import datetime
import royalnet.commands as rc
import royalnet.utils as ru
import logging
from royalnet.backpack import tables as rbt

from .abstract.linker import LinkerCommand

from ..tables import Steam, FiorygiTransaction
from ..types import Updatable

log = logging.getLogger(__name__)


class SteampoweredCommand(LinkerCommand):
    name: str = "steampowered"

    description: str = "Connetti e visualizza informazioni sul tuo account di Steam!"

    syntax: str = "{url_profilo}"

    def __init__(self, interface: rc.CommandInterface):
        super().__init__(interface)
        self._api = steam.webapi.WebAPI(self.token())

    def token(self):
        return self.config["steampowered"]["token"]

    async def get_updatables_of_user(self, session, user: rbt.User) -> List[Steam]:
        return user.steam

    async def get_updatables(self, session) -> List[Steam]:
        return await ru.asyncify(session.query(self.alchemy.get(Steam)).all)

    async def create(self,
                     session,
                     user: rbt.User,
                     args: rc.CommandArgs,
                     data: Optional[rc.CommandData] = None) -> Optional[Steam]:
        url = args.joined()
        steamid64 = await self._call(steam.steamid.steam64_from_url, url)
        if steamid64 is None:
            raise rc.InvalidInputError("Quel link non è associato ad alcun account Steam.")
        response = await self._call(self._api.ISteamUser.GetPlayerSummaries_v2, steamids=steamid64)
        r = response["response"]["players"][0]
        steam_account = self.alchemy.get(Steam)(
            user=user,
            _steamid=int(steamid64),
            persona_name=r["personaname"],
            profile_url=r["profileurl"],
            avatar=r["avatarfull"],
            primary_clan_id=r["primaryclanid"],
            account_creation_date=datetime.datetime.fromtimestamp(r["timecreated"])
        )

        await FiorygiTransaction.spawn_fiorygi(
            data=data,
            user=user,
            qty=1,
            reason="aver collegato a Royalnet il proprio account di League of Legends"
        )

        session.add(steam_account)
        return steam_account

    async def update(self, session, obj: Steam, change: Callable[[str, Any], Awaitable[None]]):
        response = await self._call(self._api.ISteamUser.GetPlayerSummaries_v2, steamids=obj.steamid.as_64)
        r = response["response"]["players"][0]
        obj.persona_name = r["personaname"]
        obj.profile_url = r["profileurl"]
        obj.avatar = r["avatar"]
        obj.primary_clan_id = r["primaryclanid"]
        obj.account_creation_date = datetime.datetime.fromtimestamp(r["timecreated"])
        response = await self._call(self._api.IPlayerService.GetSteamLevel_v1, steamid=obj.steamid.as_64)
        obj.account_level = response["response"]["player_level"]
        response = await self._call(self._api.IPlayerService.GetOwnedGames_v1,
                                    steamid=obj.steamid.as_64,
                                    include_appinfo=False,
                                    include_played_free_games=True,
                                    include_free_sub=False,
                                    appids_filter=None)
        obj.owned_games_count = response["response"]["game_count"]
        if response["response"]["game_count"] >= 0:
            obj.most_played_game_2weeks = sorted(response["response"]["games"], key=lambda g: -g.get("playtime_2weeks", 0))[0]["appid"]
            obj.most_played_game_forever = sorted(response["response"]["games"], key=lambda g: -g.get("playtime_forever", 0))[0]["appid"]

    async def on_increase(self, session, obj: Updatable, attribute: str, old: Any, new: Any) -> None:
        pass

    async def on_unchanged(self, session, obj: Updatable, attribute: str, old: Any, new: Any) -> None:
        pass

    async def on_decrease(self, session, obj: Updatable, attribute: str, old: Any, new: Any) -> None:
        pass

    async def on_first(self, session, obj: Updatable, attribute: str, old: None, new: Any) -> None:
        pass

    async def on_reset(self, session, obj: Updatable, attribute: str, old: Any, new: None) -> None:
        pass

    def describe(self, obj: Steam):
        return f"ℹ️ [url={obj.profile_url}]{obj.persona_name}[/url]\n" \
               f"[b]Level {obj.account_level}[/b]\n" \
               f"\n" \
               f"Owned games: [b]{obj.owned_games_count}[/b]\n" \
               f"Most played 2 weeks: [url=https://store.steampowered.com/app/{obj.most_played_game_2weeks}]{obj.most_played_game_2weeks}[/url]\n" \
               f"Most played forever: [url=https://store.steampowered.com/app/{obj.most_played_game_forever}]{obj.most_played_game_forever}[/url]\n" \
               f"\n" \
               f"SteamID32: [c]{obj.steamid.as_32}[/c]\n" \
               f"SteamID64: [c]{obj.steamid.as_64}[/c]\n" \
               f"SteamID2: [c]{obj.steamid.as_steam2}[/c]\n" \
               f"SteamID3: [c]{obj.steamid.as_steam3}[/c]\n" \
               f"\n" \
               f"Created on: [b]{obj.account_creation_date}[/b]\n"

    async def _call(self, method, *args, **kwargs):
        log.debug(f"Calling {method}")
        try:
            return await ru.asyncify(method, *args, **kwargs)
        except Exception as e:
            raise rc.ExternalError("\n".join(e.args).replace(self.token(), "HIDDEN"))
